from pathlib import Path
from six.moves.urllib.parse import urlunparse
from six.moves.urllib.error import HTTPError
from six.moves.urllib.request import urlretrieve
import logging


logger = logging.getLogger(__name__)


class HiRISE_URL(object):
    """Manage HiRISE URLs.

    Provide a storage path as calculated from above objects and put together the full
    URL to the HiRISE product.

    Parameters
    ----------
    product_path : str or pathlib.Path
        Storage path to the product


    """
    initurl = ('https://hirise-pds.lpl.arizona.edu/PDS/RDR/'
               'ESP/ORB_011400_011499/ESP_011491_0985/ESP_'
               '011491_0985_RED.LBL')
    scheme = 'https'
    netloc = 'hirise-pds.lpl.arizona.edu'
    pdspath = Path('/PDS')

    def __init__(self, product_path, params=None, query=None, fragment=None):
        self.product_path = product_path
        self.params = params
        self.query = query
        self.fragment = fragment

    @property
    def path(self):
        path = self.pdspath / self.product_path
        return str(path)

    @property
    def url(self):
        return urlunparse([self.scheme, self.netloc, self.path,
                           self.params, self.query, self.fragment])


class OBSERVATION_ID(object):
    """Manage HiRISE observation ids.

    For example PSP_003092_0985.

    `phase` is set to PSP for orbits < 11000, no setting required.

    Parameters
    ----------
    obsid : str, optional
        One can optionally also create an 'empty' OBSERVATION_ID object and set the
        properties accordingly to create a new obsid.
    """

    def __init__(self, obsid=None):
        if obsid is not None:
            phase, orbit, targetcode = obsid.split('_')
            self._orbit = int(orbit)
            self._targetcode = targetcode
        else:
            self._orbit = None
            self._targetcode = None

    @property
    def orbit(self):
        return str(self._orbit).zfill(6)

    @orbit.setter
    def orbit(self, value):
        if value > 999999:
            raise ValueError("Orbit cannot be larger than 999999")
        self._orbit = value

    @property
    def targetcode(self):
        return self._targetcode

    @targetcode.setter
    def targetcode(self, value):
        if len(str(value)) != 4:
            raise ValueError('Targetcode must be exactly 4 characters.')
        self._targetcode = value

    @property
    def phase(self):
        return 'PSP' if int(self.orbit) < 11000 else 'ESP'

    def __str__(self):
        return '{}_{}_{}'.format(self.phase, self.orbit, self.targetcode)

    def __repr__(self):
        return self.__str__()

    @property
    def s(self):
        return self.__str__()

    def get_upper_orbit_folder(self):
        '''
        get the upper folder name where the given orbit folder is residing on the
        hisync server
        '''
        lower = int(self.orbit) // 100 * 100
        return "_".join(["ORB", str(lower).zfill(6), str(lower + 99).zfill(6)])

    @property
    def storage_path_stem(self):
        s = "{phase}/{orbitfolder}/{obsid}".format(phase=self.phase,
                                                   orbitfolder=self.get_upper_orbit_folder(),
                                                   obsid=self.s)
        return s


class PRODUCT_ID(object):
    """Manage storage paths for HiRISE RDR products (also EXTRAS.)

    Attributes `jp2_path` and `label_path` get you the official RDR product,
    with `kind` steering if you get the COLOR or the RED product.
    All other properties go to the RDR/EXTRAS folder.

    Parameters
    ----------
    initstr : str, optional

    Note
    ----
    The "PDS" part of the path is handled in the HiRISE_URL class.

    """
    kinds = ['RED', 'BG', 'IR', 'COLOR', 'IRB', 'MIRB', 'MRGB', 'RGB']

    @classmethod
    def from_path(cls, path):
        path = Path(path)
        return cls(path.stem)

    def __init__(self, initstr=None):
        if initstr is not None:
            tokens = initstr.split('_')
            self._obsid = OBSERVATION_ID('_'.join(tokens[:3]))
            try:
                self.kind = tokens[3]
            except IndexError:
                self._kind = None
        else:
            self._kind = None

    @property
    def obsid(self):
        return self._obsid

    @obsid.setter
    def obsid(self, value):
        self._obsid = OBSERVATION_ID(value)

    @property
    def kind(self):
        return self._kind

    @kind.setter
    def kind(self, value):
        if value not in self.kinds:
            raise ValueError("kind must be in {}".format(self.kinds))
        self._kind = value

    def __str__(self):
        return "{}_{}".format(self.obsid, self.kind)

    def __repr__(self):
        return self.__str__()

    @property
    def s(self):
        return self.__str__()

    @property
    def storage_stem(self):
        return '{}/{}'.format(self.obsid.storage_path_stem, self.s)

    @property
    def label_fname(self):
        return '{}.LBL'.format(self.s)

    @property
    def label_path(self):
        return 'RDR/' + self.storage_stem + '.LBL'

    def _make_url(self, obj):
        path = getattr(self, f"{obj}_path")
        return HiRISE_URL(path).url

    def __getattr__(self, item):
        tokens = item.split('_')
        try:
            if tokens[-1] == 'url':
                return self._make_url('_'.join(tokens[:-1]))
        except IndexError:
            raise ValueError(f"No attribute named '{item}' found.")

    # TODO: implement general self.obj_url for all paths.

    @property
    def jp2_fname(self):
        return self.s + '.JP2'

    @property
    def jp2_path(self):
        prefix = 'RDR/'
        postfix = ''
        if self.kind not in ['RED', 'COLOR']:
            prefix += 'EXTRAS/'
        if self.kind in ['IRB']:
            postfix = '.NOMAP'
        return prefix + self.storage_stem + postfix + ".JP2"

    @property
    def nomap_jp2_path(self):
        if self.kind in ['RED', 'IRB', 'RGB']:
            return 'RDR/EXTRAS/' + self.storage_stem + '.NOMAP.JP2'
        else:
            raise AttributeError("No NOMAP exists for {}.".format(self.kind))

    @property
    def quicklook_path(self):
        if self.kind in ['COLOR', 'RED']:
            return Path('EXTRAS/RDR/') / (self.storage_stem + ".QLOOK.JP2")
        else:
            raise AttributeError("No quicklook exists for {} products.".format(self.kind))

    @property
    def abrowse_path(self):
        if self.kind in ['COLOR', 'MIRB', 'MRGB', 'RED']:
            return Path('EXTRAS/RDR/') / (self.storage_stem + '.abrowse.jpg')
        else:
            raise AttributeError("No abrowse exists for {}".format(self.kind))

    @property
    def browse_path(self):
        inset = ''
        if self.kind in ['IRB', 'RGB']:
            inset = '.NOMAP'
        if self.kind not in ['COLOR', 'MIRB', 'MRGB', 'RED', 'IRB', 'RGB']:
            raise AttributeError("No browse exists for {}".format(self.kind))
        else:
            return Path('EXTRAS/RDR/') / (self.storage_stem + inset + '.browse.jpg')

    @property
    def thumbnail_path(self):
        if self.kind in ['BG', 'IR']:
            raise AttributeError("No thumbnail exists for {}".format(self.kind))
        inset = ''
        if self.kind in ['IRB', 'RGB']:
            inset = '.NOMAP'
        return Path('EXTRAS/RDR/') / (self.storage_stem + inset + '.thumb.jpg')

    @property
    def nomap_thumbnail_path(self):
        if self.kind in ['RED', 'IRB', 'RGB']:
            return Path('EXTRAS/RDR') / (self.storage_stem + '.NOMAP.thumb.jpg')
        else:
            raise AttributeError("No NOMAP thumbnail exists for {}".format(self.kind))

    @property
    def nomap_browse_path(self):
        if self.kind in ['RED', 'IRB', 'RGB']:
            return Path('EXTRAS/RDR') / (self.storage_stem + '.NOMAP.browse.jpg')

    @property
    def edr_storage_stem(self):
        return 'EDR/' + self.storage_stem


class SOURCE_PRODUCT_ID(object):
    """Manage SOURCE_PRODUCT_ID.

    Example
    -------
    'PSP_003092_0985_RED4_0'
    """

    red_ccds = ['RED' + str(i) for i in range(10)]
    ir_ccds = ['IR10', 'IR11']
    bg_ccds = ['BG12', 'BG13']
    ccds = red_ccds + ir_ccds + bg_ccds

    def __init__(self, spid=None, saveroot=None):
        if spid is not None:
            tokens = spid.split('_')
            obsid = '_'.join(tokens[:3])
            ccd = tokens[3]
            color, ccdno = self._parse_ccd(ccd)
            self.pid = PRODUCT_ID('_'.join([obsid, color]))
            self.ccd = ccd
            self.channel = tokens[4]
            self.saveroot = saveroot
        else:
            self.pid = None
            self._channel = None
            self._ccd = None

    def __getattr__(self, value):
        return getattr(self.pid, value)

    def _parse_ccd(self, value):
        sep = 2 if value[:2] in PRODUCT_ID.kinds else 3
        return value[:sep], value[sep:]

    @property
    def channel(self):
        return self._channel

    @channel.setter
    def channel(self, value):
        if int(value) not in [0, 1]:
            raise ValueError("channel must be in [0, 1]")
        self._channel = value

    @property
    def ccd(self):
        return self._ccd

    @ccd.setter
    def ccd(self, value):
        if value not in self.ccds:
            raise ValueError("CCD value must be in {}.".format(self.ccds))
        self._ccd = value
        if self.pid is not None:
            self.pid.color = self.color

    @property
    def color(self):
        return self._parse_ccd(self.ccd)[0]

    @property
    def ccdno(self):
        offset = len(self.color)
        return self.ccd[offset:]

    def __str__(self):
        return "{}: {}{}_{}".format(self.__class__.__name__, self.pid, self.ccdno, self.channel)

    def __repr__(self):
        return self.__str__()

    @property
    def s(self):
        return "{}{}_{}".format(self.pid, self.ccdno, self.channel)

    @property
    def fname(self):
        return self.s + '.IMG'

    @property
    def local_cube(self):
        return self.local_path.with_suffix('.cub')

    @property
    def fpath(self):
        return Path(self.pid.edr_storage_stem).parent / self.fname

    @property
    def furl(self):
        hiurl = HiRISE_URL(self.fpath)
        return hiurl.url

    @property
    def stitched_cube_name(self):
        return f"{self.pid.obsid.s}_{self.ccd}.cub"

    @property
    def local_path(self):
        savepath = self.saveroot / str(self.obsid) / self.fname
        return savepath

    def download(self, overwrite=False):
        savepath = self.local_path
        if savepath.exists() and not overwrite:
            logger.warning("File exists and I'm not allowed to overwrite:"
                           " %s", savepath)
            return
        savepath.parent.mkdir(parents=True, exist_ok=True)
        logger.info(f"Downloading\n{self.furl}\nto\n{savepath}")
        try:
            urlretrieve(self.furl, str(savepath))
        except HTTPError as e:
            logger.error(e.__str__())


class RED_PRODUCT_ID(SOURCE_PRODUCT_ID):
    def __init__(self, obsid, ccdno, channel, **kwargs):
        self.ccds = self.red_ccds
        super().__init__('{}_RED{}_{}'.format(obsid, ccdno, channel),
                         **kwargs)


class IR_PRODUCT_ID(SOURCE_PRODUCT_ID):
    def __init__(self, obsid, ccdno, channel):
        self.ccds = self.ir_ccds
        super().__init__('{}_IR{}_{}'.format(obsid, ccdno, channel))



import numpy as np


def rebin(a, newshape):
    """Rebin an array to a new shape."""
    assert len(a.shape) == len(newshape)

    slices = [slice(0, old, float(old) / new) for old, new in zip(a.shape, newshape)]
    coordinates = np.mgrid[slices]
    indices = coordinates.astype('i')  # choose the biggest smaller integer index
    print(len(indices), indices.max())
    return a[tuple(indices)]


def rebin_factor(a, newshape):
    '''Rebin an array to a new shape.
    newshape must be a factor of a.shape.
    '''
    assert len(a.shape) == len(newshape)
    assert not np.sometrue(np.mod(a.shape, newshape))

    slices = [slice(None, None, old / new) for old, new in zip(a.shape, newshape)]
    return a[slices]



import subprocess
from pathlib import Path

import click
from imageio import imread
from six.moves.urllib.error import HTTPError
from six.moves.urllib.request import urlretrieve

from .products import PRODUCT_ID, RED_PRODUCT_ID, HiRISE_URL

try:
    from pptx import Presentation
except ImportError:
    PPTX_INSTALLED = False
else:
    PPTX_INSTALLED = True


def hirise_dropbox():
    return Path.home() / 'Dropbox' / 'data' / 'hirise'


def labels_root():
    return hirise_dropbox() / 'labels'


def browse_root():
    return hirise_dropbox() / 'browse'


def get_rdr_some_label(kind, obsid):
    """Download `some` PRODUCT_ID label for `obsid`.

    Note, that the RED channel is also called the B&W channel on the HiRISE website.

    Parameters
    ----------
    kind : {'RED', 'COLOR'}
        String that determines the kind of color looking for.
    obsid : str
        HiRISE obsid in the standard form of ESP_012345_1234

    Returns
    -------
    None
        Storing the label file in the `labels_root` folder.

    """
    pid = PRODUCT_ID(obsid)
    pid.kind = kind
    savepath = labels_root() / Path(pid.label_fname)
    savepath.parent.mkdir(exist_ok=True)
    print("Downloading\n", pid.label_url, 'to\n', savepath)
    try:
        urlretrieve(pid.label_url, str(savepath))
    except HTTPError as e:
        print(e)


def get_rdr_red_label(obsid):
    """Download the RED PRODUCT_ID label for `obsid`.

    Note
    ----
    The RED channel is also called the B&W channel on the HiRISE website.

    Parameters
    ----------
    obsid : str
        HiRISE obsid in the standard form of ESP_012345_1234

    Returns
    -------
    None
        Storing the label file in the `labels_root` folder.
    """
    get_rdr_some_label('RED', obsid)


def get_rdr_color_label(obsid):
    """Download the COLOR PRODUCT_ID label for `obsid`.

    Parameters
    ----------
    obsid : str
        HiRISE obsid in the standard form of ESP_012345_1234

    Returns
    -------
    None
        Storing the label file in the `labels_root` folder.
    """
    get_rdr_some_label('COLOR', obsid)


def download_product(prodid_path, saveroot=None):
    if saveroot is None:
        saveroot = hirise_dropbox()
    else:
        saveroot = Path(saveroot)
        if not saveroot.is_absolute():
            saveroot = hirise_dropbox() / saveroot

    saveroot.mkdir(exist_ok=True)
    url = HiRISE_URL(prodid_path)
    savepath = saveroot / prodid_path.name
    print("Downloading\n", url.url, 'to\n', savepath)
    try:
        urlretrieve(url.url, str(savepath))
    except HTTPError as e:
        print(e)
    return savepath


def download_RED_product(obsid, ccdno, channel, saveroot=None, overwrite=False):
    if saveroot is None:
        saveroot = hirise_dropbox()
    else:
        saveroot = Path(saveroot)
        if not saveroot.is_absolute():
            saveroot = hirise_dropbox() / saveroot

    pid = RED_PRODUCT_ID(obsid, ccdno, channel)
    savepath = saveroot / obsid / pid.fname
    savepath.parent.mkdir(parents=True, exist_ok=True)

    # if file already is there:
    if savepath.exists() and not overwrite:
        return savepath

    print("Downloading\n", pid.furl, '\nto\n', savepath)
    try:
        urlretrieve(pid.furl, str(savepath))
    except HTTPError as e:
        print(e)
    return savepath


def download_browse_product(obsid, kind='RED', annotated=True, saveroot='.', overwrite=False):
    """Download a browse product from HiRISE website.

    By default, store it in current path.

    Parameters
    ----------
    obsid : str
        HiRISE observation ID
    kind : {'RED','COLOR'}, optional
        String indicating the kind of required browse product. Default: 'RED'
    annotated : bool, optional
        Switch to control if the annotated version is required. Default: True
    saveroot : str, pathlib.Path, optional
        Path in where the browse product is being stored. Default: '.'
    overwrite : bool, optional
        Boolean switch to control if an existing path should be overwritten. Default: False
    """
    pid = PRODUCT_ID(f"{obsid}_{kind}")
    if annotated is True:
        url = pid.abrowse_url
    else:
        url = pid.browse_url
    saveroot = Path(saveroot)
    savepath = saveroot / Path(url).name
    savepath.parent.mkdir(parents=True, exist_ok=True)

    if savepath.exists() and not overwrite:
        print("Path exists. No download required.", savepath)
        return savepath

    print("Downloading\n", url, '\nto\n', savepath)
    try:
        urlretrieve(url, str(savepath))
    except HTTPError as e:
        print(e)
    return savepath


@click.command()
@click.argument('obsid')
def get_and_display_browse_product(obsid):
    path = download_browse_product(obsid)
    subprocess.run(["open", "-a", "Preview", path])


def create_browse_presentation(obsids, savename='obsid_browse_images', **kwargs):
    if not PPTX_INSTALLED:
        print("You need to install the `pptx` module for this to work.")
        return
    savepath = str(savename) + '.pptx'
    prs = Presentation()

    pic_left = int(prs.slide_width * 0.0)
    pic_top = int(prs.slide_height * 0.0)

    blank_slide_layout = prs.slide_layouts[6]

    for obsid in obsids[0:]:
        imgpath = str(download_browse_product(obsid, **kwargs))
        slide = prs.slides.add_slide(blank_slide_layout)
        img = imread(imgpath)
        ratio = img.shape[0] / img.shape[1]
        if ratio < 1:
            pic_width = int(prs.slide_width)
            pic_height = int(pic_width * img.shape[0] / img.shape[1])
        else:
            pic_height = int(prs.slide_height)
            pic_width = int(pic_height / ratio)
        slide.shapes.add_picture(imgpath, pic_left, pic_top, pic_width, pic_height)

    prs.save(savepath)




import matplotlib.pyplot as plt
import pandas as pd
import pvl

from .downloads import hirise_dropbox


def get_rdr_index_names():
    lblfile = hirise_dropbox() / 'RDRCUMINDEX.LBL'
    label = pvl.load(str(lblfile))
    table = label['RDR_INDEX_TABLE']
    names = []
    for item in table:
        second = item[1]
        if isinstance(second, pvl.PVLObject):
            names.append(second['NAME'])
    return names


def get_rdr_index():
    names = get_rdr_index_names()
    indexfile = hirise_dropbox() / 'RDRCUMINDEX.TAB'
    df = pd.read_csv(indexfile, names=names)
    return df


class PolyPlotter(object):
    """For plotting the outline of HiRISE RDR polygons.
    """
    # note, that last point needs to repeat 1 point to close polygon
    lat_indices = ['CORNER1_LATITUDE',
                   'CORNER2_LATITUDE',
                   'CORNER3_LATITUDE',
                   'CORNER4_LATITUDE',
                   'CORNER1_LATITUDE']
    lon_indices = ['CORNER1_LONGITUDE',
                   'CORNER2_LONGITUDE',
                   'CORNER3_LONGITUDE',
                   'CORNER4_LONGITUDE',
                   'CORNER1_LONGITUDE']

    def __init__(self):
        self.df = get_rdr_index()
        self.df.set_index('PRODUCT_ID', inplace=True)

        print("Loaded RDR index file.")

    def plot_prodid(self, product_id, ax=None, **kwargs):
        if ax is None:
            _, ax = plt.subplots()
        ax.plot(self.df.loc[product_id][self.lon_indices],
                self.df.loc[product_id][self.lat_indices], **kwargs)



import pvl


class HiRISE_Label(object):

    def __init__(self, fname):
        self.label = pvl.load(str(fname))

    @property
    def binning(self):
        return self.label['INSTRUMENT_SETTING_PARAMETERS']['MRO:BINNING']

    @property
    def lines(self):
        return self.label['UNCOMPRESSED_FILE']['IMAGE']['LINES']

    @property
    def line_samples(self):
        return self.label['UNCOMPRESSED_FILE']['IMAGE']['LINE_SAMPLES']

    @property
    def l_s(self):
        return self.label['VIEWING_PARAMETERS']['SOLAR_LONGITUDE'].value

    @property
    def map_scale(self):
        return self.label['IMAGE_MAP_PROJECTION']['MAP_SCALE'].value



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from setuptools import setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "Click>=6.0",
    # TODO: put package requirements here
]

test_requirements = [
    # TODO: put package test requirements here
]


setup(
    name="hirise_tools",
    version="0.7.0",
    description="Tools to work with MRO's HiRISE camera data.",
    long_description=readme + "\n\n" + history,
    author="K.-Michael Aye",
    author_email="kmichael.aye@gmail.com",
    url="https://github.com/michaelaye/pyrise",
    packages=["pyrise",],
    package_dir={"pyrise": "pyrise"},
    entry_points={
        "console_scripts": [
            "pyrise = pyrise.cli:main",
            "abrowse = pyrise.downloads:get_and_display_browse_product",
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="ISC license",
    zip_safe=False,
    keywords=["HiRISE", "MRO", "NASA", "PDS",],
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Topic :: Utilities",
    ],
    test_suite="tests",
    test_require=test_requirements,
)
